@@ -1,6 +1,6 @@
 /*
   SDL_image:  An example image loading library for use with SDL
-  Copyright (C) 1997-2016 Sam Lantinga <slouken@libsdl.org>
+  Copyright (C) 1997-2023 Sam Lantinga <slouken@libsdl.org>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -23,10 +23,6 @@
 
 /* This is a Targa image file loading framework */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-
 #include "SDL_endian.h"
 
 #include "SDL_image.h"
@@ -301,7 +297,7 @@ SDL_Surface *IMG_LoadTGA_RW(SDL_RWops *src)
         } else {
             SDL_RWread(src, dst, w * bpp, 1);
         }
-#if SDL_BYTEORDER == SDL_LIL_ENDIAN
+#if SDL_BYTEORDER == SDL_BIG_ENDIAN
         if (bpp == 2) {
             /* swap byte order */
             int x;
@@ -327,6 +323,9 @@ error:
 }
 
 #else
+#if _MSC_VER >= 1300
+#pragma warning(disable : 4100) /* warning C4100: 'op' : unreferenced formal parameter */
+#endif
 
 /* dummy TGA load routine */
 SDL_Surface *IMG_LoadTGA_RW(SDL_RWops *src)
